#property strict

#include <EventBase.mqh>

class INDICATOR; // 

class BUFFER : public EVENTBASE
{
private:
  double Visual[]; //    

  static int BuffersTotal;
  static int BuffersInteractiveTotal;

//       ,    
  static const BUFFER* PrevBuffer;
  static int CountBuffers;

  static void MyChartRedraw( const BUFFER &This );

  const INDICATOR* IndicatorOver; // ,   

  int TimeToVisual; //     .

  ENUM_LINE_STYLE Style;
  int Width;
  color  Color;

  string Name;

  void _SetIndexStyle( const bool Redraw = FALSE ) const
  {
    if (this.MT4IndicatorBuffer)
    {
      ::PlotIndexSetInteger(this.Index, PLOT_LINE_COLOR, this.Color);
      ::PlotIndexSetInteger(this.Index, PLOT_DRAW_TYPE, this.Type);
      ::PlotIndexSetInteger(this.Index, PLOT_LINE_STYLE, this.Style);

      if (Redraw)
        ::ChartRedraw();
    }
    return;
  }

  void FullCopyBuffer( void )
  {
    if (this.Interactive)
    {
      ::ArrayCopy(Visual, Buffer);

      BUFFER::MyChartRedraw(this);
    }

    this.TimeToVisual = 0;

    return;
  }

  void InteractiveLine( const double Step )
  {
    const int Size = ::ArraySize(this.Visual);

    for (int i = 0; i < Size; i++)
      if ((this.Visual[i] != EMPTY_VALUE) && (this.Buffer[i] != EMPTY_VALUE))
        this.Visual[i] += (this.Buffer[i] - this.Visual[i]) * Step;

    return;
  }

  static int GetArrayPos( const double &Array[], int BeginPos = 0, const double Value = EMPTY_VALUE )
  {
    const int Size = ::ArraySize(Array);

    for (int Pos = BeginPos; Pos < Size; Pos++)
      if (Array[Pos] != Value)
        return(Pos);

    return(Size);
  }

  bool CorrectIndex( const int iIndex ) const
  {
    return(((iIndex >= 0) && (iIndex < ::ArraySize(this.Buffer))));
  }

  void InteractiveSection( const double Step )
  {
    const int Size = ::ArraySize(this.Visual);
    int i1 = -1, j1 = -1;
    int i2 = this.GetArrayPos(this.Buffer);
    int j2 = this.GetArrayPos(this.Visual);

    while (this.CorrectIndex(i2) && this.CorrectIndex(j2))
    {
      if (i2 == j2)
      {
        this.Visual[j2] += (this.Buffer[i2] - this.Visual[j2]) * Step;

        i1 = i2;
        i2 = this.GetArrayPos(this.Buffer, i2 + 1);

        j1 = j2;
        j2 = this.GetArrayPos(this.Visual, j2 + 1);
      }
      else if (i2 > j2)
      {
        if (CorrectIndex(i1))
        {
          const double Value = this.Buffer[i1] + (this.Buffer[i2] - this.Buffer[i1]) * (j2 - i1) / (i2 - i1);

          this.Visual[j2] += (Value - this.Visual[j2]) * Step;
        }

        j1 = j2;
        j2 = this.GetArrayPos(this.Visual, j2 + 1);
      }
      else // (i2 < j2)
      {
        if (CorrectIndex(j1))
        {
          const double Value = this.Visual[j1] + (this.Visual[j2] - this.Visual[j1]) * (i2 - j1) / (j2 - j1);

          this.Visual[i2] = Value + (this.Buffer[i2] - Value) * Step;
        }

        i1 = i2;
        i2 = this.GetArrayPos(this.Buffer, i2 + 1);
      }
    }

    return;
  }

  //  -   -
  const BUFFER* GetSecondZigZag( void ) const
  {
    if ((this.Type != DRAW_ZIGZAG) || ::CheckPointer(this.IndicatorOver) == POINTER_INVALID)
      return(NULL);

    const BUFFER* Pointer = ::GetPointer(this);
    const BUFFER* Res = NULL;
    const int Amount = this.IndicatorOver.GetAmountBuffers();
    int Pos;
    uchar Sum = 0;

    for (Pos = 0; Pos < Amount; Pos++)
    {
      const BUFFER* CurrentBuffer = this.IndicatorOver[Pos];

      if (CurrentBuffer.Type == this.Type)
      {
        if (CurrentBuffer == Pointer)
          break;

        Res = CurrentBuffer;
        Sum++;
      }
    }

    if (((Sum & 0x01) == 0) || (Pos >= Amount))
    {
      Res = NULL;

      for (Pos = Pos + 1; Pos < Amount; Pos++)
      {
        const BUFFER* CurrentBuffer = this.IndicatorOver[Pos];

        if (CurrentBuffer.Type == this.Type)
        {
          Res = CurrentBuffer;

          break;
        }
      }
    }

    return(Res);
  }

  void InteractiveZigZag( const double Step )
  {
//    const BUFFER* Buffer2 = this.GetSecondZigZag(); //  -

    this.FullCopyBuffer(); //     DRAW_ZIGZAG

    return;
  }

  void InteractiveStep( void )
  {
    if (this.TimeToVisual == 0)
      return;

    const int Interval = (int)this.GetTimerInterval();

    if ((this.TimeToVisual < Interval) || (Interval < 0))
      this.FullCopyBuffer();
    else if (this.Interactive && (Interval > 0))
    {
      const double Step = (double)Interval / this.TimeToVisual;

      switch (this.Type)
      {
        case DRAW_LINE:
        case DRAW_HISTOGRAM:
        case DRAW_ARROW:
          InteractiveLine(Step);

          break;

        case DRAW_SECTION:
          InteractiveSection(Step);

          break;
        case DRAW_ZIGZAG:
          InteractiveZigZag(Step);
      }

      this.TimeToVisual -= Interval;

      BUFFER::MyChartRedraw(this);
    }

    return;
  }

public:
  double Buffer[]; //    (  )

  const int Index;
  const int Type;

  const bool MT4IndicatorBuffer;

  const bool Interactive;

  void SetName( const string sName = "" )
  {
    if (this.Name != sName)
    {
      this.Name = sName;

      if (this.MT4IndicatorBuffer)
      #ifdef __MQL5__
        ::PlotIndexSetString(this.Index, ::PLOT_LABEL, this.Name);
      #endif
      #ifdef __MQL4__
        ::SetIndexLabel(this.Index, this.Name);
      #endif
    }

    return;
  }

  void SetWidth( const int iWidth = EMPTY, const bool Redraw = TRUE )
  {
    if (this.Width != iWidth)
    {
      this.Width = iWidth;

      this._SetIndexStyle(Redraw);
    }

    return;
  }

  virtual void SetScale( const bool Redraw = TRUE )
  {
    long Scale;

    if (::ChartGetInteger(0, CHART_SCALE, 0, Scale))
    {
      const int iWidth = 9 / (1 << (5 - (int)Scale));

      this.SetWidth(iWidth, Redraw);
    }

    return;
  }

  void SetStyle( const ENUM_LINE_STYLE iStyle = EMPTY, const bool Redraw = TRUE )
  {
    if (this.Style != iStyle)
    {
      this.Style = iStyle;

      this._SetIndexStyle(Redraw);
    }

    return;
  }

  void SetColor( const color cColor = clrNONE, const bool Redraw = TRUE )
  {
    if (this.Color != cColor)
    {
      this.Color = cColor;

      this._SetIndexStyle(Redraw);
    }

    return;
  }

  void SetEmptyValue( const int Start = 0, int iCount = WHOLE_ARRAY )
  {
    if (iCount == WHOLE_ARRAY)
      iCount = ::ArraySize(this.Buffer) - Start;

    if (iCount > 0)
      ::ArrayFill(this.Buffer, Start, iCount, EMPTY_VALUE);

    return;
  }

  string GetName( void ) const
  {
    return(this.Name);
  }

  int GetWidth( void ) const
  {
    return(this.Width);
  }

  ENUM_LINE_STYLE GetStyle( void ) const
  {
    return(this.Style);
  }

  color GetColor( void ) const
  {
    return(this.Color);
  }

  // MT4- 
  BUFFER( const int iType, const ENUM_LINE_STYLE eStyle = EMPTY,
          const int iWidth = EMPTY, const color cColor = clrNONE, const string sName = "", const bool AddEvent = FALSE, const bool bInteractive = FALSE ) :
          EVENTBASE(AddEvent), Type(iType), Style(eStyle), Width(iWidth), Color(cColor),
          Name(""), Index(BUFFER::BuffersTotal), MT4IndicatorBuffer(TRUE), Interactive(AddEvent ? bInteractive : FALSE), TimeToVisual(0), IndicatorOver(NULL)
  {
    // WARNING MT5-ERROR: https://www.mql5.com/ru/forum/6343/page557#comment_2386968
    if (this.Interactive)
    {
      ::SetIndexBuffer(this.Index, this.Visual);

      #ifdef __MQL4__
        const int TmpIndex = indicator_buffers + this.Index; //     #property

        ::IndicatorBuffers(TmpIndex + 1);
      #endif

      #ifdef __MQL5__
        //  MT5  ::IndicatorBuffers,     indicator_buffers!!!
        const int TmpIndex = indicator_plots + this.Index; //     #property
      #endif

      ::SetIndexBuffer(TmpIndex, this.Buffer, INDICATOR_CALCULATIONS);
    }
    else
      ::SetIndexBuffer(this.Index, this.Buffer);

    this._SetIndexStyle();

    this.SetName(sName);

    this.BuffersTotal++;

    if (this.Interactive)
      this.BuffersInteractiveTotal++;
  }

  //  MT4- 
  BUFFER( const string sName = "", const bool AddEvent = FALSE ) : EVENTBASE(AddEvent),
          Type(EMPTY), Style(EMPTY), Width(EMPTY), Color(clrNONE), Name(sName), Index(EMPTY),
          MT4IndicatorBuffer(FALSE), Interactive(FALSE), TimeToVisual(0), IndicatorOver(NULL)
  {
  }

  ~BUFFER( void )
  {
    this.BuffersTotal--;

    if (this.Interactive)
      this.BuffersInteractiveTotal--;
  }

  double operator []( const int Pos ) const
  {
    return(this.Buffer[Pos]);
  }

  int GetAmount( void ) const
  {
    return(::ArraySize(this.Buffer));
  }

  static int GetMT4IndicatorBuffersTotal( void )
  {
    return(BUFFER::BuffersTotal);
  }

// -  
  virtual void EventTimer( void )
  {
    InteractiveStep();

    return;
  }

  void ToVisual( const INDICATOR &iIndicatorOver, const int iTimeToVisual = 0 )
  {
    this.IndicatorOver = ::GetPointer(iIndicatorOver);

    if (this.TimeToVisual == 0)
      this.TimeToVisual = iTimeToVisual;

    if (this.TimeToVisual < (int)this.GetTimerInterval())
      this.FullCopyBuffer();

    return;
  }
};

static int BUFFER::BuffersTotal = 0;
static int BUFFER::BuffersInteractiveTotal;

//       ,    
static const BUFFER* BUFFER::PrevBuffer = NULL;
static int BUFFER::CountBuffers = 0;

static void BUFFER::MyChartRedraw( const BUFFER &This )
{
  if (BUFFER::PrevBuffer != ::GetPointer(This))
  {
    BUFFER::PrevBuffer = ::GetPointer(This);

    BUFFER::CountBuffers++;
  }

  if (BUFFER::CountBuffers == BUFFER::BuffersInteractiveTotal)
  {
    ::ChartRedraw();

    BUFFER::PrevBuffer = NULL;

    BUFFER::CountBuffers = 0;
  }

  return;
}